/*
 * ext_all.c - Dump entire configuration at external interface
 *
 * Written 2001 by Werner Almesberger 
 * Copyright 2001 Network Robots
 */

/*
 * This is experimental code and will change. Probably a lot.
 *
 * Known problems: @@@
 * - should allow simple configurations using other qdiscs than dsmark
 * - does not support qdiscs shared by multiples classes
 * - gets confused if dumping more than one qdisc tree (ingress/egress) per
 *   session
 * - does not take into account that only part of the dsmark class may be used
 *   for marking
 * - does not take into account that no dsmark class may exist for marking
 *   (in fact, should look at parameters to decide whether a change is
 *   actually requested)
 * - does not report inner qdiscs changing skb->tc_index
 */


#include <inttypes.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
#include "error.h"
#include "param.h"
#include "tree.h"
#include "filter.h"
#include "qdisc.h"
#include "if.h"
#include "iflib.h"
#include "ext.h"
#include "ext_all.h"


int dump_all; /* dump all qdiscs and filters (experimental) */


static const QDISC *_root; /* @@@@ HACK */


/* -------------------- skb->tc_index to class mapping --------------------- */


static void print_class_n(FILE *file,const QDISC *qdisc,uint16_t class,
  int *first)
{
    fprintf(file,"%c%u:%u",*first ? ' ' : ',',(unsigned) qdisc->number,
      (unsigned) class);
    *first = 0;
}


static void print_class(FILE *file,const CLASS *class,int *first)
{
    print_class_n(file,class->parent.qdisc,class->number,first);
}


static void gred_path(FILE *file,const QDISC *qdisc,uint16_t index,int first)
{
    const CLASS *class;

    param_get(qdisc->params);
    for (class = qdisc->classes; class; class = class->next)
	if (class->number == (index & 0xf)) {
	    print_class(file,class,&first);
	    return;
	}
    print_class_n(file,qdisc,prm_default_index.v,&first);
}


static void no_filters_at_classes(const CLASS *class)
{
    while (class) {
	if (class->filters)
	    yyerror("dump_all currently does not support filters at classes");
	no_filters_at_classes(class->child);
	class = class->next;
    }
}


static CLASS *find_class(CLASS *classes,uint32_t number)
{
    CLASS *class;

    for (class = classes; class; class = class->next) {
	if (class->number == number) return class;
	if (class->child) {
	    CLASS *found;

	    found = find_class(class->child,number);
	    if (found) return found;
	}
    }
    return NULL;
}


static void find_path(FILE *file,const QDISC *qdisc,uint16_t index,int first);


static void filter_path(FILE *file,const QDISC *qdisc,uint16_t index,int first)
{
    const FILTER *filters = qdisc->filters;
    const FILTER *filter;


    if (!filters) {
	if (qdisc->dsc->flags & QDISC_HAS_CLASSES)
	    yyerrorf("don't know how to select classes of \"%s\"",
	      qdisc->dsc->name);
	return;
    }
    for (filter = filters; filter; filter = filter->next) {
	const ELEMENT *element;

	if (filter->dsc != &tcindex_dsc)
	    yyerrorf("dump_all supports only \"tcindex\", not \"%s\"",
	      filter->dsc->name);
	for (element = filter->elements; element; element = element->next)
	    if (element->police)
		yyerror("dump_all does not support policing at tcindex");
    }
    /*
     * Note: filter.c ensures that filters are sorted by priority, with the
     * hightest priority (lowest number) first.
     */
    for (filter = filters; filter; filter = filter->next) {
	const ELEMENT *element;
	CLASS *class;
	int shift,mask;
	unsigned value;

	param_load(filter->params,filter->dsc->filter_param,
	  filter->dsc->filter_param);
	shift = prm_shift.present ? prm_shift.v : 0;
	mask = prm_mask.present ? prm_mask.v : 0xffff;
	value = (index & mask) >> shift;
	debugf("at qdisc %u: (index 0x%x & mask 0x%x) >> shift %d -> 0x%x",
	  (unsigned) qdisc->number,(unsigned) index,mask,shift,value);
	for (element = filter->elements; element; element = element->next) {
	    param_push();
	    param_load(element->params,filter->dsc->filter_param,
	      filter->dsc->element_param);
	    if ((index & mask) >> shift == element->number) {
		class = element->parent.class;
		print_class(file,class,&first);
		if (class->qdisc) find_path(file,class->qdisc,index,first);
		param_pop();
		return;
	    }
	    param_pop();
	}
	if (!prm_fall_through.present || !prm_fall_through.v) continue;
	print_class_n(file,qdisc,value,&first);
	class = find_class(qdisc->classes,value);
	if (class && class->qdisc) find_path(file,class->qdisc,index,first);
	return;
    }
    yyerrorf("dump_all found no matching choice for value 0x%x in tcindex",
      (unsigned) index);
}


static void find_path(FILE *file,const QDISC *qdisc,uint16_t index,int first)
{
    no_filters_at_classes(qdisc->classes);
    if (qdisc->dsc == &gred_dsc) gred_path(file,qdisc,index,first);
    else filter_path(file,qdisc,index,first);
}


int dump_all_decision(FILE *file,uint16_t class)
{
    const CLASS *c;

    if (!dump_all) return 0;
    for (c = _root->classes; c; c = c->next)
	if (c->qdisc) {
	    int one = 1;

	    fprintf(file," class");
	    print_class_n(file,_root,class,&one);
	    find_path(file,c->qdisc,class,0);
	    return 1;
	}
    return 0;
}


/* ------------------- Dumps qdisc and class definitions ------------------- */


static void print_pragma(FILE *file,PARAM *params)
{
    const DATA_LIST *pragma;

    if (!prm_present(params,&prm_pragma)) return;
    fprintf(file," pragma");
    for (pragma = prm_data(params,&prm_pragma).u.list; pragma;
      pragma = pragma->next)
	fprintf(file," %s",pragma->ref->u.string);
}


static PARAM_DSC pseudo_prm_parent ={
    .id = "parent",
};

static PARAM_DSC pseudo_prm_qdisc = {
    .id = "qdisc",
};


static void class_special_fn(FILE *file,const PARAM_DSC *param,void *arg)
{
    const CLASS *class = arg;

    if (param == &pseudo_prm_parent && class->parent.class)
	fprintf(file," parent %d",(int) class->parent.class->number);
    if (param == &pseudo_prm_qdisc)
	fprintf(file," qdisc %d",(int) class->qdisc->number);
}


static void dump_class(FILE *file,const CLASS *class)
{
    const QDISC *qdisc = class->parent.qdisc;
    static const PARAM_DSC *default_special[] = {
	&pseudo_prm_parent,
	&prm_pragma,
	&prm_priomap, /* CBQ  */
	NULL
    };
    const PARAM_DSC **special = default_special;
    void (*special_fn)(FILE *file,const PARAM_DSC *param,void *arg);

    special_fn = class_special_fn;
    fprintf(file,"class %u =",(unsigned) class->number);
    /* @@@ <HACKS>*/
    if (qdisc->dsc == &gred_dsc) {
	static const PARAM_DSC *gred_special[] = {
	    &pseudo_prm_parent,
	    &prm_default,
	    &prm_pragma,
	    NULL
	};

	special = gred_special;
    }
    if (class->qdisc) {
	static const PARAM_DSC *class_qdisc_special[] = {
	    &pseudo_prm_parent,
	    &pseudo_prm_qdisc,
	    &prm_pragma,
	    &prm_priomap, /* CBQ  */
	    NULL
	};

	special = class_qdisc_special;
    }
    /* @@@ </HACKS>*/
    param_print(file,qdisc->dsc->qdisc_param,qdisc->dsc->class_param,
      qdisc->dsc->class_param,special,special_fn,(void *) class);
    print_pragma(file,class->params);
    fprintf(file,"\n");
}


static void dump_classes(FILE *file,const CLASS *class)
{
    while (class) {
	param_push();
	param_load(class->params,class->parent.qdisc->dsc->qdisc_param,
	  class->parent.qdisc->dsc->class_param);
	dump_class(file,class);
	dump_classes(file,class->child);
	param_pop();
	class = class->next;
    }
}


static void dump_this_qdisc(FILE *file,const QDISC *qdisc);


static void dump_qdiscs_at_classes(FILE *file,const CLASS *class)
{
    while (class) {
	if (class->qdisc) dump_this_qdisc(file,class->qdisc);
	dump_qdiscs_at_classes(file,class->child);
	class = class->next;
    }
}


static void gred_default(FILE *file,const PARAM_DSC *param,void *arg)
{
    const QDISC *qdisc = arg;
    const CLASS *class;

    if (param == &prm_pragma) return;
    for (class = qdisc->classes; class; class = class->next)
	if (prm_present(class->params,&prm_default)) {
	    fprintf(file," default %d",(int) class->number);
	    return;
	}
    abort();
}


static void dump_this_qdisc(FILE *file,const QDISC *qdisc)
{
    static const PARAM_DSC *hide_pragma[] = {
	&prm_pragma,
	&prm_priomap, /* CBQ (see note below) p*/
	NULL
    };
	/*
	 * Note concerning CBQ: param_print does not understand that priomap
	 * is not defaulted from qdisc to class, so it already hides priomap,
	 * although for the wrong reason. Including it here anyway ensures
	 * that priomap will remain hidden when param_print is fixed.
	 */
    const PARAM_DSC **special = hide_pragma;
    void (*special_fn)(FILE *file,const PARAM_DSC *param,void *arg) = NULL;

    dump_qdiscs_at_classes(file,qdisc->classes);
    fprintf(file,"qdisc %u = %s",(unsigned) qdisc->number,qdisc->dsc->name);
    param_load(qdisc->params,qdisc->dsc->qdisc_param,qdisc->dsc->qdisc_param);
    /* @@@ <HACKS>*/
    if (qdisc->dsc == &gred_dsc) {
	static const PARAM_DSC *insert_default[] = {
	    &prm_default,
	    NULL
	};

	special = insert_default;
	special_fn = gred_default;
    }
    if (qdisc->dsc == &prio_dsc) fprintf(file," bands 3");
    /* @@@ </HACKS>*/
    param_print(file,qdisc->dsc->qdisc_param,qdisc->dsc->class_param,
      qdisc->dsc->qdisc_param,special,special_fn,(void *) qdisc);
    print_pragma(file,qdisc->params);
    fprintf(file,"\n");
    dump_classes(file,qdisc->classes);
}


/* ------------------------------------------------------------------------- */


static void dump_callback(FILE *file,void *user)
{
    QDISC *root = user;

    do_dump_global(root,file);
    dump_this_qdisc(file,root);
    if (root->filters) {
	root->if_expr = iflib_combine(root);
	do_dump_if_ext(root->filters,file);
    }
}


static void no_root_class_filters(const CLASS *classes)
{
    const CLASS *class;

    for (class = classes; class; class = class->next) {
	if (class->filters)
	    yyerrorf("dump_all: classes of root qdisc may not have filters");
	no_root_class_filters(class->child);
    }
}


static void no_dsmark_constraints(const CLASS *classes)
{
    const CLASS *class;

    for (class = classes; class; class = class->next) {
	if (class->qdisc &&
	  (class->qdisc->dsc->flags & (QDISC_HAS_CLASSES | QDISC_HAS_FILTERS)))
	    yyerror("dump_all: if not dsmark, only root qdisc may have "
	      "classes");
	no_dsmark_constraints(class->child);
    }
}


void dump_all_qdisc(QDISC *root,const char *target)
{
    if (!root) yyerror("dump_all: no root qdisc");
    _root = root;
    if (root->filters) {
	if (root->filters->next) yyerror("dump_all: only one filter allowed");
	if (root->filters->dsc != &if_dsc)
	    yyerrorf("dump_all: filter must be \"if\", not \"%s\"",
	      root->filters->dsc->name);
    }
    no_root_class_filters(root->classes);
    if (root->dsc != &dsmark_dsc) no_dsmark_constraints(root->classes);
    ext_build(target,root,NULL,dump_callback,root);
}
